@@ -1,8 +1,9 @@
 /*
     sredird: RFC 2217 compliant serial port redirector
-    Version 2.0.0, 14 January 2002
-    Copyright (C) 1999, 2000, 2001, 2002 InfoTecna s.r.l.
-    Copyright (C) 2001  Trustees of Columbia University in the City of New York
+    Version 2.2.0, 09 July 2003
+    Copyright (C) 1999 - 2003 InfoTecna s.r.l.
+    Copyright (C) 2001, 2002 Trustees of Columbia University
+    in the City of New York
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,7 +19,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    To contact the autor:
+    To contact the authors:
 
       Denis Sbragion
       InfoTecna
@@ -63,6 +64,33 @@
       . Track the state of the telnet negotiations
 
       . Add support for BINARY mode translations
+
+   Fixed in 2.1.0:
+
+      . GetPortFlowControl should return 1 to indicate NO FLOW CONTROL 
+        instead of 0.  
+  
+      . The Cisco IOS hack should become activated only if set by command-
+        line option [-i].
+      
+      . Changed the order of checks in the EscWriteChar function for slightly
+        better performance
+
+   Fixed in 2.2.0:
+
+      Mario Viara
+
+      Email: mario@viara.info
+
+      . Fixed set port data size now work with 5 6 7 8 bits.
+      . Add version in get signature.
+
+      Russell Coker <russell@coker.com.au> 
+
+      . Many minor changes and code cleanup
+      
+      For other important changes from Russell Coker see the README file.
+        
 */
 
 /* Return NoError, which is 0, on success */
@@ -89,7 +117,8 @@
 #include <netinet/tcp.h>
 
 /* Version id */
-#define SRedirdVersionId "Version 2.0.0, 14 January 2002"
+#define VersionId "2.2.0"
+#define SRedirdVersionId "Version " VersionId ", 09 July 2003"
 
 /* Locking constants */
 #define LockOk 0
@@ -170,6 +199,9 @@
 /* Standard boolean definition */
 typedef enum { False, True } Boolean;
 
+/* Cisco IOS bug compatibility */
+Boolean CiscoIOSCompatible = False;
+
 /* Buffer structure */
 typedef
   struct
@@ -186,7 +218,7 @@ static char * LockFileName;
 /* Complete device file pathname */
 static char * DeviceName;
 
-/* True when the device ha been opened */
+/* True when the device has been opened */
 Boolean DeviceOpened = False;
 
 /* Device file descriptor */
@@ -225,8 +257,10 @@ static unsigned char ModemStateMask = ((unsigned char) 255);
 /* Line state mask set by the client */
 static unsigned char LineStateMask = ((unsigned char) 0);
 
+#ifdef COMMENT
 /* Current status of the line control lines */
 static unsigned char LineState = ((unsigned char) 0);
+#endif
 
 /* Current status of the modem control lines */
 static unsigned char ModemState = ((unsigned char) 0);
@@ -272,10 +306,10 @@ unsigned char GetFromBuffer(BufferType * B);
 
 /* Generic log function with log level control. Uses the same log levels
 of the syslog(3) system call */
-void LogMsg(int LogLevel, char * Msg);
+void LogMsg(int LogLevel, const char * const Msg);
 
 /* Try to lock the file given in LockFile as pid LockPid using the classical
-  HDB (ASCII) file locking scheme */
+HDB (ASCII) file locking scheme */
 int HDBLockFile(char * LockFile, pid_t LockPid);
 
 /* Remove the lock file created with HDBLockFile */
@@ -285,10 +319,10 @@ void HDBUnlockFile(char * LockFile, pid_t LockPid);
 void ExitFunction(void);
 
 /* Function called on many signals */
-void SignalFunction(int Act);
+void SignalFunction(int unused);
 
 /* Function called on break signal */
-void BreakFunction(int Act);
+void BreakFunction(int unused);
 
 /* Retrieves the port speed from PortFd */
 unsigned long int GetPortSpeed(int PortFd);
@@ -325,37 +359,37 @@ void SetPortFlowControl(int PortFd,unsigned char How);
 void SetPortSpeed(int PortFd, unsigned long BaudRate);
 
 /* Send the signature Sig to the client */
-void SendSignature(BufferType *, char * Sig);
+void SendSignature(BufferType * B, char * Sig);
 
 /* Write a char to SockFd performing IAC escaping */
-void EscWriteChar(BufferType *, unsigned char C);
+void EscWriteChar(BufferType * B, unsigned char C);
 
 /* Redirect char C to PortFd checking for IAC escape sequences */
-void EscRedirectChar(BufferType *, BufferType *, int PortFd, unsigned char C);
+void EscRedirectChar(BufferType * SockB, BufferType * DevB, int PortFd, unsigned char C);
 
 /* Send the specific telnet option to SockFd using Command as command */
-void SendTelnetOption(BufferType *, unsigned char Command, char Option);
+void SendTelnetOption(BufferType * B, unsigned char Command, char Option);
 
 /* Send a string to SockFd performing IAC escaping */
-void SendStr(BufferType *, char * Str);
+void SendStr(BufferType * B, char * Str);
 
 /* Send the baud rate BR to SockFd */
-void SendBaudRate(BufferType *, unsigned long int BR);
+void SendBaudRate(BufferType * B, unsigned long int BR);
 
 /* Send the flow control command Command */
-void SendCPCFlowCommand(BufferType *, unsigned char Command);
+void SendCPCFlowCommand(BufferType * B, unsigned char Command);
 
 /* Send the CPC command Command using Parm as parameter */
-void SendCPCByteCommand(BufferType *, unsigned char Command, unsigned char Parm);
+void SendCPCByteCommand(BufferType * B, unsigned char Command, unsigned char Parm);
 
 /* Handling of COM Port Control specific commands */
-void HandleCPCCommand(BufferType *, int PortFd, unsigned char * Command, size_t CSize);
+void HandleCPCCommand(BufferType * B, int PortFd, unsigned char * Command, size_t CSize);
 
 /* Common telnet IAC commands handling */
-void HandleIACCommand(BufferType *, int PortFd, unsigned char * Command, size_t CSize);
+void HandleIACCommand(BufferType * B, int PortFd, unsigned char * Command, size_t CSize);
 
 /* Write a buffer to SockFd with IAC escaping */
-void EscWriteBuffer(BufferType *, unsigned char * Buffer, unsigned int BSize);
+void EscWriteBuffer(BufferType * B, unsigned char * Buffer, unsigned int BSize);
 
 /* initialize Telnet State Machine */
 void InitTelnetStateMachine(void)
@@ -418,15 +452,15 @@ unsigned char GetFromBuffer(BufferType * B)
   }
 
 /* Generic log function with log level control. Uses the same log levels
-  of the syslog(3) system call */
-void LogMsg(int LogLevel, char * Msg)
+of the syslog(3) system call */
+void LogMsg(int LogLevel, const char * const Msg)
   {
     if (LogLevel <= MaxLogLevel)
       syslog(LogLevel,Msg);
   }
 
 /* Try to lock the file given in LockFile as pid LockPid using the classical
-  HDB (ASCII) file locking scheme */
+HDB (ASCII) file locking scheme */
 int HDBLockFile(char * LockFile, pid_t LockPid)
   {
     pid_t Pid;
@@ -449,7 +483,7 @@ int HDBLockFile(char * LockFile, pid_t LockPid)
             if (N <= 0)
               {
                 /* Emtpy lock file or error: may be another application
-                  was writing its pid in it */
+                was writing its pid in it */
                 sprintf(LogStr,"Can't read pid from lock file %s.",LockFile);
                 LogMsg(LOG_NOTICE,LogStr);
 
@@ -574,17 +608,27 @@ void ExitFunction(void)
   }
 
 /* Function called on many signals */
-void SignalFunction(int Act)
+void SignalFunction(int unused)
   {
+    /* Just to avoid compilation warnings */
+    /* There's no performance penalty in doing this 
+    because this function is almost never called */
+    unused = unused;
+    
     /* Same as the exit function */
     ExitFunction();
   }
 
 /* Function called on break signal */
 /* Unimplemented yet */
-void BreakFunction(int Act)
+void BreakFunction(int unused)
   {
 #ifndef COMMENT
+    /* Just to avoid compilation warnings */
+    /* There's no performance penalty in doing this 
+    because this function is almost never called */
+    unused = unused;
+
     /* Same as the exit function */
     ExitFunction();
 #else /* COMMENT */
@@ -720,7 +764,7 @@ unsigned char GetPortStopSize(int PortFd)
   }
 
 /* Retrieves the flow control status, including DTR and RTS status,
-  from PortFd */
+from PortFd */
 unsigned char GetPortFlowControl(int PortFd, unsigned char Which)
   {
     struct termios PortSettings;
@@ -739,7 +783,7 @@ unsigned char GetPortFlowControl(int PortFd, unsigned char Which)
             return((unsigned char) 2);
           if (PortSettings.c_cflag & CRTSCTS)
             return((unsigned char) 3);
-          return((unsigned char) 0);
+          return((unsigned char) 1);
         break;
 
         /* BREAK State  */
@@ -753,17 +797,17 @@ unsigned char GetPortFlowControl(int PortFd, unsigned char Which)
         /* DTR Signal State */
         case 7:
           if (MLines & TIOCM_DTR)
-            return((unsigned char) 9);
-          else
             return((unsigned char) 8);
+          else
+            return((unsigned char) 9);
         break;
 
         /* RTS Signal State */
         case 10:
           if (MLines & TIOCM_RTS)
-            return((unsigned char) 12);
-          else
             return((unsigned char) 11);
+          else
+            return((unsigned char) 12);
         break;
 
         /* Com Port Flow Control Setting (inbound) */
@@ -780,7 +824,7 @@ unsigned char GetPortFlowControl(int PortFd, unsigned char Which)
             return((unsigned char) 2);
           if (PortSettings.c_cflag & CRTSCTS)
             return((unsigned char) 3);
-          return((unsigned char) 0);
+          return((unsigned char) 1);
         break;
       }
   }
@@ -839,8 +883,8 @@ void SetPortDataSize(int PortFd, unsigned char DataSize)
       }
 
     tcgetattr(PortFd,&PortSettings);
-    PortSettings.c_cflag = PortSettings.c_cflag &
-      ((PortSettings.c_cflag & ~CSIZE) | PDataSize);
+    PortSettings.c_cflag &= ~CSIZE;
+    PortSettings.c_cflag |= PDataSize & CSIZE;
     tcsetattr(PortFd,TCSADRAIN,&PortSettings);
   }
 
@@ -937,7 +981,7 @@ void SetPortFlowControl(int PortFd,unsigned char How)
         break;
         /* BREAK State OFF */
         case 6:
-          /* should not send another break */
+          /* Should not send another break */
           /* tcsendbreak(PortFd,0); */
           BreakSignaled = False;
         break;
@@ -1074,7 +1118,7 @@ void EscWriteChar(BufferType * B, unsigned char C)
     if (C == TNIAC)
       AddToBuffer(B,C);
     else
-      if (!tnstate[TN_TRANSMIT_BINARY].is_will && Last == 0x0D && C != 0x0A)
+      if (C != 0x0A && !tnstate[TN_TRANSMIT_BINARY].is_will && Last == 0x0D)
         AddToBuffer(B,0x00);
     AddToBuffer(B,C);
 
@@ -1097,7 +1141,7 @@ void EscRedirectChar(BufferType * SockB, BufferType * DevB, int PortFd, unsigned
             IACEscape = IACReceived;
           else
             if (!tnstate[TN_TRANSMIT_BINARY].is_do && C == 0x00 && Last == 0x0D)
-              /* swallow the NUL after a CR if not receiving BINARY */
+              /* Swallow the NUL after a CR if not receiving BINARY */
               break;
             else
               AddToBuffer(DevB,C);
@@ -1269,7 +1313,7 @@ void SendStr(BufferType * B, char * Str)
 /* Send the baud rate BR to Buffer */
 void SendBaudRate(BufferType *B, unsigned long int BR)
   {
-    unsigned char C, *p;
+    unsigned char *p;
     unsigned long int NBR;
     int i;
 
@@ -1279,8 +1323,8 @@ void SendBaudRate(BufferType *B, unsigned long int BR)
     AddToBuffer(B,TNSB);
     AddToBuffer(B,TNCOM_PORT_OPTION);
     AddToBuffer(B,TNASC_SET_BAUDRATE);
-    p = (char *)&NBR;
-    for (i = 0;i < sizeof(NBR);i++)
+    p = (unsigned char *) &NBR;
+    for (i = 0;i < (int) sizeof(NBR);i++)
       EscWriteChar(B,p[i]);
     AddToBuffer(B,TNIAC);
     AddToBuffer(B,TNSE);
@@ -1289,8 +1333,6 @@ void SendBaudRate(BufferType *B, unsigned long int BR)
 /* Send the flow control command Command */
 void SendCPCFlowCommand(BufferType *B, unsigned char Command)
   {
-    unsigned char C;
-
     AddToBuffer(B,TNIAC);
     AddToBuffer(B,TNSB);
     AddToBuffer(B,TNCOM_PORT_OPTION);
@@ -1307,8 +1349,6 @@ void SendCPCFlowCommand(BufferType *B, unsigned char Command)
 /* Send the CPC command Command using Parm as parameter */
 void SendCPCByteCommand(BufferType *B, unsigned char Command, unsigned char Parm)
   {
-    unsigned char C;
-
     AddToBuffer(B,TNIAC);
     AddToBuffer(B,TNSB);
     AddToBuffer(B,TNCOM_PORT_OPTION);
@@ -1337,7 +1377,7 @@ void HandleCPCCommand(BufferType *SockB, int PortFd, unsigned char * Command, si
           if (CSize == 6)
             {
               /* Void signature, client is asking for our signature */
-              sprintf(SigStr,"SRedird %s",DeviceName);
+              sprintf(SigStr,"SRedird %s %s",VersionId,DeviceName);
               SendSignature(SockB,SigStr);
               sprintf(LogStr,"Sent signature: %s",SigStr);
               LogMsg(LOG_INFO,LogStr);
@@ -1345,7 +1385,7 @@ void HandleCPCCommand(BufferType *SockB, int PortFd, unsigned char * Command, si
           else
             {
               /* Received client signature */
-              strncpy(SigStr,&Command[4],CSize - 6);
+              strncpy(SigStr,(char *) &Command[4],CSize - 6);
               sprintf(LogStr,"Received client signature: %s",SigStr);
               LogMsg(LOG_INFO,LogStr);
             }
@@ -1475,12 +1515,13 @@ void HandleCPCCommand(BufferType *SockB, int PortFd, unsigned char * Command, si
                 SetPortFlowControl(PortFd,Command[4]);
 
                 /* Flow control status confirmation */
-                if (Command[4] >= 13 && Command[4] <=16)
+                if (CiscoIOSCompatible && Command[4] >= 13 && Command[4] <=16)
                   /* INBOUND not supported separately.
                     Following the behavior of Cisco ISO 11.3
                   */
                   FlowControl = 0;
                 else
+                  /* Return the actual port flow control settings */
                   FlowControl = GetPortFlowControl(PortFd,0);
 
                 SendCPCByteCommand(SockB,TNASC_SET_CONTROL,FlowControl);
@@ -1727,6 +1768,26 @@ void EscWriteBuffer(BufferType * B, unsigned char * Buffer, unsigned int BSize)
         }
   }
 
+void Usage(void)
+{
+    /* Write little usage information */
+    puts("sredird: RFC 2217 compliant serial port redirector");
+    puts(SRedirdVersionId);
+    puts("This program should be run only by the inetd superserver");
+    puts("Usage: sredird [-i] <loglevel> <device> <lockfile> [pollingterval]");
+    puts("-i indicates Cisco IOS Bug compatibility");
+    puts("Poll interval is in milliseconds, default is 100, "
+          "0 means no polling");
+
+    /* Same on the system log */
+    LogMsg(LOG_ERR,"sredird: RFC 2217 compliant serial port redirector.");
+    LogMsg(LOG_ERR,SRedirdVersionId);
+    LogMsg(LOG_ERR,"This program should be run only by the inetd superserver.");
+    LogMsg(LOG_ERR,"Usage: sredird [-i] <loglevel> <device> <lockfile> [pollingterval]");
+    LogMsg(LOG_ERR,"-i indicates Cisco IOS Bug compatibility");
+    LogMsg(LOG_ERR,"Poll interval is in milliseconds, default is 100, 0 means no polling.");
+}
+
 /* Main function */
 int main(int argc, char * argv[])
   {
@@ -1769,11 +1830,12 @@ int main(int argc, char * argv[])
     /* Generic socket parameter */
     int SockParm;
 
-    /* Out buffer ticks count */
-    clock_t OutBTicks = times(NULL);
-
     /* Out buffer clock ticks limit */
     clock_t MaxBTicks;
+    
+    /* Optional argument processing indexes */
+    int argi = 1;
+    int i;
 
     /* Open the system log */
     openlog("sredird",LOG_PID,LOG_USER);
@@ -1781,29 +1843,47 @@ int main(int argc, char * argv[])
     /* Check the command line argument count */
     if (argc < 4)
       {
-        /* Write little usage information */
-        puts("sredird: RFC 2217 compliant serial port redirector");
-        puts(SRedirdVersionId);
-        puts("This program should be run only by the inetd superserver");
-        puts("Usage: sredird <loglevel> <device> <lockfile> [pollingterval]");
-        puts("Poll interval is in milliseconds, default is 100, "
-          "0 means no polling");
-
-        /* Same on the system log */
-        LogMsg(LOG_ERR,"sredird: RFC 2217 compliant serial port redirector.");
-        LogMsg(LOG_ERR,SRedirdVersionId);
-        LogMsg(LOG_ERR,"This program should be run only by the inetd superserver.");
-        LogMsg(LOG_ERR,"Usage: sredird <loglevel> <device> <lockfile> [pollingterval]");
-        LogMsg(LOG_ERR,"Poll interval is in milliseconds, default is 100, 0 means no polling.");
-
+        Usage();
         return(Error);
       }
 
+    /* Process optional switch arguments */
+    for (argi = 1;argv[argi][0] == '-' && argi < argc;argi++) 
+      {
+        i = 1;
+        while (argv[argi][i])
+          {
+            switch (argv[argi][i++])
+              {
+                /* Cisco IOS compatibility */
+                case 'i':
+                  if (CiscoIOSCompatible)
+                    {
+                      /* Already set */
+                      Usage();
+                      return(Error);
+                    }
+                  else
+                    CiscoIOSCompatible = True;
+                break;
+                
+                default:
+                  Usage();
+                  return(Error);
+                break;
+              }
+          }
+      }
+
     /* Sets the log level */
-    MaxLogLevel = atoi(argv[1]);
+    MaxLogLevel = atoi(argv[argi++]);
+
+    /* Gets device and lock file names */
+    DeviceName = argv[argi++];
+    LockFileName = argv[argi++];
 
     /* Retrieve the polling interval */
-    if (argc == 5)
+    if (argc == argi + 1)
       {
         BTimeout.tv_sec = 0;
         BTimeout.tv_usec = atol(argv[4]) * 1000;
@@ -1833,10 +1913,6 @@ int main(int argc, char * argv[])
     sprintf(LogStr,"Polling interval (ms): %u",(unsigned int) (BTimeout.tv_usec / 1000));
     LogMsg(LOG_INFO,LogStr);
 
-    /* Gets device and lock file names */
-    DeviceName = argv[2];
-    LockFileName = argv[3];
-
     /* Register exit and signal handler functions */
     atexit(ExitFunction);
     signal(SIGHUP,SignalFunction);
